package http

import (
	"bytes"
	"encoding/json"
	"fmt"
	"io"
	"net/http/httptest"
	"testing"
	"time"

	"github.com/gogo/protobuf/jsonpb"
	"github.com/gogo/protobuf/proto"
	"github.com/stretchr/testify/assert"

	"bitbucket.org/orus-io/go-xbus"
	"bitbucket.org/orus-io/go-xbus/api"
	"bitbucket.org/orus-io/xbus/fullenv_tests"
	"bitbucket.org/orus-io/xbus/testutils"
)

func must(t *testing.T, err error) {
	if err != nil {
		t.Fatal(err)
	}
}

type LoggingReader struct {
	reader io.Reader
	tb     testing.TB
}

func (r LoggingReader) Read(p []byte) (n int, err error) {
	n, err = r.reader.Read(p)
	r.tb.Logf("Just read %s", string(p))
	return
}

func setupEnv(t *testing.T, filename string) (env *fullenv.FullEnv, gw *Gateway) {
	defer func() {
		if gw == nil {
			env.Close()
		}
	}()
	logger := testutils.GetTestLogger(t)
	options, err := fullenv.OptionsFromFile(filename)
	options.Logger = logger
	must(t, err)
	env = fullenv.NewFullEnv("xbus-http-emitter", options)
	must(t, env.Init())

	must(t, env.Startup())

	coptions := xbus.COptions{
		Name: "gw",
		Type: api.Account_GATEWAY,
	}
	client := xbus.NewClient(coptions, logger)
	must(t, client.CreateKey(false))
	must(t, client.CreateCSR("localhost", "xbus", "xbus", false))
	must(t, client.Connect())
	s, err := client.Register()
	must(t, err)
	if s != api.Registration_PENDING {
		t.Fatal(fmt.Sprintf("Invalid state: %s", s))
	}
	must(t, env.XbusCtl.Run("account", "accept", "gw"))
	s, err = client.Register()
	must(t, err)
	if s != api.Registration_ACCEPTED {
		t.Fatal(fmt.Sprintf("Invalid state: %s", s))
	}
	client.Close()
	coptions = client.GetOptions()
	gw = NewGateway(coptions, logger)
	return
}

func TestEmitter(t *testing.T) {
	for _, encoding := range []string{"json", "proto"} {
		t.Run(encoding, func(t *testing.T) {
			env, gateway := setupEnv(t, "emitter-test.yaml")
			defer env.Close()

			apiKey, err := env.XbusCtl.RunReadLastLine("account", "renew-apikey", "outerclient")
			must(t, err)

			var buf io.Reader
			if encoding == "json" {
				buf = bytes.NewBufferString(`{
		"output": "default",
		"envelope": {
			"event_ids": ["53131466-4e2e-11e7-916e-af2b1e6d242d"],
			"events": [{
				"id": "53131466-4e2e-11e7-916e-af2b1e6d242d",
				"type": "test",
				"ItemCount": 1,
				"Items": ["QSBtZXNzYWdl"]
			}]
		}
	}`)
			} else {
				b, err := proto.Marshal(&api.OutputRequest{
					Output: "default",
					Envelope: &api.Envelope{
						EventIDs: []api.UUID{
							api.MustUUIDFromString("53131466-4e2e-11e7-916e-af2b1e6d242d"),
						},
						Events: []api.Event{
							{ID: api.MustUUIDFromString("53131466-4e2e-11e7-916e-af2b1e6d242d"),
								Type:      "test",
								ItemCount: 1,
								Items:     []api.Item{api.Item("A message")},
							},
						},
					},
				})
				if err != nil {
					t.Fatal(err)
				}
				buf = bytes.NewBuffer(b)
			}

			req := httptest.NewRequest("POST", "/emitter/output", buf)
			if encoding == "json" {
				req.Header.Set("Content-Type", "application/json")
			} else {
				req.Header.Set("Content-Type", "application/x-protobuf")
			}
			req.Header.Set("Xbus-API-Key", apiKey)

			w := httptest.NewRecorder()

			gateway.Router.ServeHTTP(w, req)

			if w.Result().StatusCode != 200 {
				var body = make([]byte, 1024*1024)
				t.Errorf("Invalid status: %s", w.Result().Status)
				c, err := w.Result().Body.Read(body)
				if err != nil {
					t.Fatalf("Could not read body: %s", err)
				}
				t.Fatalf("Body is: %s", string(body[:c]))
			}

			if string(w.Body.Bytes()) != "OK" {
				t.Errorf("Unexpected response: %s", w.Body.Bytes())
				return
			}

			time.Sleep(time.Second)

			assert.Equal(t, 1, len(env.Consumer.Envelopes))

			e := env.Consumer.Envelopes[0]
			assert.Equal(t, "A message", string(e.Events[0].Items[0]))
			assert.Equal(t, "53131466-4e2e-11e7-916e-af2b1e6d242d", e.Events[0].ID.String())
		})
	}
}

func TestConsumer(t *testing.T) {
	env, gateway := setupEnv(t, "consumer-test.yaml")
	defer env.Close()

	apiKey, err := env.XbusCtl.RunReadLastLine("account", "renew-apikey", "outerclient")
	must(t, err)

	req := httptest.NewRequest("GET", "/test-consumer/processrequest", nil)
	req.Header.Set("Xbus-API-Key", apiKey)
	req.Header.Set("Accept", "application/json")
	req.Header.Set("Accept-Item-Encoding", "json,text,binary")

	w := httptest.NewRecorder()

	sender := env.Emitter.Actor.OpenOutput(nil, "default", "test")
	must(t, sender.AddItems("test", api.Item("a item")))
	must(t, sender.Close())

	envStateChan, sub, err := env.Emitter.Actor.GetClient().API.NewProcessState(
		env.Emitter.Actor.ID.String(),
	).EnvelopeSubscribeChan(sender.GetID().String())
	if err != nil {
		t.Fatal(err)
	}
	defer sub.Unsubscribe()

	done := make(chan bool)

	go func() {
		gateway.Router.ServeHTTP(w, req)
		done <- true
	}()

	select {
	case <-time.After(10 * time.Second):
		t.Fatal("Timeout...")
	case <-done:
		t.Log("Request served, checking result")
	}

	if w.Result().StatusCode != 200 {
		var body = make([]byte, 1024*1024)
		t.Errorf("Invalid status: %s", w.Result().Status)
		c, err := w.Result().Body.Read(body)
		if err != nil {
			t.Fatalf("Could not read body: %s", err)
		}
		t.Fatalf("Body is: %s", string(body[:c]))
	}

	dec := json.NewDecoder(LoggingReader{w.Body, t})

	var reply api.ActorProcessRequest

	if err := jsonpb.UnmarshalNext(dec, &reply); err != nil {
		t.Fatalf("Error decoding response: %s", err)
	}
	envelope := reply.Inputs[0].Envelope
	if len(envelope.Events) != 1 {
		t.Fatalf("Expected 1 event, got %d", len(envelope.Events))
	}
	event := envelope.Events[0]
	if len(event.Items) != 1 {
		t.Fatalf("Expected 1 item, got %d", len(event.Items))
	}
	item := event.Items[0]
	assert.Equal(t, "a item", string(item))

	// TODO decode the envelope and check its items

	{
		buf := bytes.NewBuffer(nil)
		enc := json.NewEncoder(buf)
		if err := enc.Encode(map[string]interface{}{
			"context": reply.Context,
			"status":  "SUCCESS",
		}); err != nil {
			t.Fatal(err)
		}

		req := httptest.NewRequest("POST", "/test-consumer/processingend", buf)
		req.Header.Set("Xbus-API-Key", apiKey)

		w := httptest.NewRecorder()

		done := make(chan bool)

		go func() {
			gateway.Router.ServeHTTP(w, req)
			done <- true
		}()

		select {
		case <-time.After(10 * time.Second):
			t.Fatal("Timeout...")
		case <-done:
			t.Log("Request served, checking result")
		}

		if w.Result().StatusCode != 200 {
			t.Fatalf("Unexpected status code: %d\nBody: %s", w.Result().StatusCode, w.Body.String())
		}

	}
	timeout := time.After(5 * time.Second)
	for {
		select {
		case <-timeout:
			t.Errorf("Process state not updated to 'done'")
			return
		case envState := <-envStateChan:
			if envState.Status == api.Process_DONE {
				return
			}
		}
	}
}
