// Copyright 2016 Orus.io. All rights reserved.

package director

import (
	"context"
	"fmt"
	"sync"
	"time"

	"github.com/nats-io/go-nats"
	"github.com/nats-rpc/gogo-nrpc"

	"bitbucket.org/orus-io/go-xbus/api"
	"bitbucket.org/orus-io/go-xbus/tools"
	"bitbucket.org/orus-io/xbus/process"
	"bitbucket.org/orus-io/xbus/services"
	"bitbucket.org/orus-io/xbus/storage"
)

func min(a, b uint) uint {
	if a < b {
		return a
	}
	return b
}

const (
	// DirectorQueue is the name we use to queueSub to a queue with the director
	DirectorQueue = "director_queue"
)

// Options is used to create a Service
type Options struct {
	services.Options

	ActorStorage    storage.ActorStorage
	EnvelopeStorage storage.EnvelopeStorage
	LogStorage      storage.LogStorage
	ProcessManager  *process.Manager
	JobStorage      storage.JobStorage
}

// NewService returns a new Service
func (o Options) NewService() *Service {
	return &Service{
		BaseService:    services.NewBaseService("director", o.Options),
		Opts:           o,
		processManager: o.ProcessManager,
		running:        make(map[api.UUID][]*Job),
		envWatcher:     NewEnvelopeWatcher(),
	}
}

// New returns a new Service
func New(options Options) *Service {
	return options.NewService()
}

// Job ...
type Job struct {
	storage.Job
	firstattempt time.Time
	attempt      uint
	timeoutCount uint

	// If currently running, cancel will stop the job
	cancel func()
}

// Service is the director service implementation
// The director service is responsible for calling the 'Process' function
// on the actors.
// It monitors the ProcessNodeReadyEvent events and schedule jobs
// When a actor is back online or a when a process is resumed, the corresponding
// jobs are resheduled
// The director service acts alone a is not (yet) able to cooperate with other
// director instances.
type Service struct {
	services.BaseService
	Opts Options

	processManager *process.Manager

	api             *api.Client
	directorHandler *api.DirectorHandler
	actorMap        *ActorMap
	jobs            chan Job
	pushback        chan *Job

	envWatcher *EnvelopeWatcher

	ctx    context.Context
	cancel func()

	done chan struct{}

	// Jobs for which 'Process' is being called so they can be canceled
	running     map[api.UUID][]*Job
	runningLock sync.Mutex
}

func (s *Service) mainloop() {
	defer s.SafeTaskDone()

	actorQueues := make(map[api.UUID]*JobQueue)

	defer func() {
		for _, queue := range actorQueues {
			queue.Close()
		}
	}()

	setOnline := func(id api.UUID) {
		// init the queue if needed
		queue, ok := actorQueues[id]
		if !ok {
			queue = newJobQueue(
				s.Opts.JobStorage, id, 1024,
			)
			actorQueues[id] = queue
		}

		// add a worker on the queue
		// TODO for parallel-able actors, start more routine
		s.TaskAdd()
		go s.worker(queue.C())
	}

	for _, id := range s.actorMap.ListOnline() {
		setOnline(id)
	}

	for {
		select {
		case job := <-s.jobs:
			queue, ok := actorQueues[job.Node.ActorID]
			// If a queue exists, add the job to it
			if ok {
				queue.Add(&job)
			}
		case event := <-s.actorMap.Events:
			if event.Online {
				setOnline(event.ID)
			} else {
				// drop the queue worker and mark the queue for deletion
				if queue, ok := actorQueues[event.ID]; ok {
					// TODO in some case the call may happen during shutdown,
					// after the queue is closed, which crashed the server.
					// To reproduce, unsubscribe a client during a Process()
					// handling and shutdown everything.
					go queue.SendNilToOut()
					// TODO mark the queue for deletion
				}
			}
		case job := <-s.pushback:
			queue, ok := actorQueues[job.Node.ActorID]

			// If a queue exists, re-add the job to it in first position
			if ok {
				queue.PushBack(job)
			}

		case <-s.done:
			return
		}
	}
}

// Startup initialize all the subscriptions
func (s *Service) Startup() error {
	s.ctx, s.cancel = context.WithCancel(context.Background())
	s.directorHandler = api.NewDirectorHandler(
		s.ctx, s.Conn, directorServer{s},
	)
	if err := s.SubscribeAll(
		s.HandlerQueueSubscribe(s.directorHandler, DirectorQueue),
	); err != nil {
		return err
	}

	s.done = make(chan struct{})
	s.jobs = make(chan Job)
	s.pushback = make(chan *Job)

	s.api = api.NewClient(s.Conn)

	sub, err := s.api.CoreEvents.ProcessNodeReadySubscribe(
		s.handleProcessNodeReady)
	if err != nil {
		return err
	}
	s.AddSubscription(sub)

	sub, err = s.api.CoreEvents.ProcessSubscribe(s.handleProcess)
	if err != nil {
		return err
	}
	s.AddSubscription(sub)

	if m, err := NewActorMap(s.ctx, s.api, s.Log); err == nil {
		s.actorMap = m
	} else {
		return err
	}

	sub, err = s.api.CoreEvents.EnvelopeSubscribe(s.envWatcher.PushEvent)
	if err != nil {
		return err
	}
	s.AddSubscription(sub)

	s.TaskAdd()
	go s.mainloop()

	return nil
}

// Shutdown the service
func (s *Service) Shutdown() error {
	s.actorMap.Close()
	close(s.done)
	t := time.AfterFunc(5*time.Second, func() {
		s.Log.Noticef("Forcing shutdown of current calls")
		s.cancel()
	})
	err := s.BaseService.Shutdown()
	t.Stop()
	close(s.jobs)
	s.jobs = nil
	return err
}

type directorServer struct {
	*Service
}

func (s directorServer) Log(
	ctx context.Context, actorSID string, req api.ActorLogRequest,
) (err error) {
	actorID, err := api.UUIDFromString(actorSID)
	if err != nil {
		return
	}

	if req.EnvelopeID != nil && !req.EnvelopeID.IsZero() {
		_, err = s.Opts.ProcessManager.Get(req.Context.ProcessID)
		if err != nil {
			return
		}
	}

	var (
		envID, processID api.UUID
		nodeID           string
	)
	if req.EnvelopeID != nil {
		envID = *req.EnvelopeID
	}
	if ctx := req.Context; ctx != nil {
		nodeID = ctx.NodeID
		processID = ctx.ProcessID
	}
	for _, msg := range req.Messages {
		s.Opts.LogStorage.Log(api.LogEntry{
			EnvelopeID: envID,
			ActorID:    actorID,
			NodeID:     nodeID,
			ProcessID:  processID,
			Message:    msg,
		})
	}
	return
}

func (s directorServer) ProcessingEnd(
	ctx context.Context, actorSID string, req api.ActorProcessingState,
) (err error) {
	actorID, err := api.UUIDFromString(actorSID)
	if err != nil {
		return
	}

	p, err := s.Opts.ProcessManager.Get(req.Context.ProcessID)
	if err != nil {
		return
	}

	if req.Status != api.ActorProcessingState_SUCCESS &&
		req.Status != api.ActorProcessingState_ERROR {
		return fmt.Errorf("Invalid status: %s", req.Status)
	}

	for _, msg := range req.Messages {
		s.Opts.LogStorage.Log(api.LogEntry{
			ActorID:   actorID,
			NodeID:    req.Context.NodeID,
			ProcessID: req.Context.ProcessID,
			Message:   msg,
		})
	}

	s.Opts.JobStorage.SetState(storage.JobID{
		ProcessID: req.Context.ProcessID,
		Node: storage.ProcessNode{
			ActorID: actorID,
			NodeID:  req.Context.NodeID,
		},
	}, storage.JobState{
		Detached: true,
		Status:   req.Status,
	})
	p.ActorProcessingEnd(actorID, req.Context.NodeID, req.Status)
	return
}

func (s *Service) handleProcessNodeReady(event api.ProcessNodeReadyEvent) {
	sJob := storage.Job{
		JobID: storage.JobID{
			ProcessID: event.ProcessID,
			Node: storage.ProcessNode{
				ActorID: event.ActorID,
				NodeID:  event.NodeID,
			},
		},
	}
	s.Opts.JobStorage.NewJob(sJob.JobID, sJob.JobState)

	s.jobs <- Job{Job: sJob}
}

func (s *Service) handleProcess(event api.CoreEvent) {
	s.TaskAdd()
	defer s.SafeTaskDone()

	if event.Name == "status" {
		if event.ProcessOldStatus == api.Process_PAUSED &&
			event.Process.Status == api.Process_RUNNING {

			for _, j := range s.Opts.JobStorage.ListJobByProcess(event.Process.ID) {
				s.Log.Noticef("Rescheduling processings for %s", j.Node)
				s.jobs <- Job{Job: j}
			}
		} else if event.Process.Status == api.Process_ERROR {
			// we need to cancel ongoing processing if any
			s.cancelProcessJobs(event.Process.ID)
		}
	}
}

func (s *Service) worker(jobs <-chan *Job) {
	defer s.SafeTaskDone()

	for {
		select {
		case job := <-jobs:
			if job == nil {
				s.Log.Debugf("Worker: Exiting")
				return
			}
			s.handleJob(job)
		case <-s.ctx.Done():
			return
		}
	}
}

func (s *Service) setJobRunning(job *Job) (context.Context, func()) {
	s.runningLock.Lock()
	defer s.runningLock.Unlock()

	ctx, cancel := context.WithCancel(s.ctx)
	job.cancel = cancel

	processJobList, _ := s.running[job.ProcessID]
	s.running[job.ProcessID] = append(processJobList, job)

	return ctx, cancel
}

func (s *Service) unsetJobRunning(job *Job) {
	s.runningLock.Lock()
	defer s.runningLock.Unlock()

	processJobList, _ := s.running[job.ProcessID]
	for i, j := range processJobList {
		if j == job {
			processJobList = append(
				processJobList[:i], processJobList[i+1:]...)
			break
		}
	}
	if len(processJobList) == 0 {
		delete(s.running, job.ProcessID)
	} else {
		s.running[job.ProcessID] = processJobList
	}
}

func (s *Service) cancelProcessJobs(id api.UUID) {
	s.runningLock.Lock()
	defer s.runningLock.Unlock()

	processJobList, _ := s.running[id]
	for _, job := range processJobList {
		job.cancel()
	}
}

func (s *Service) handleJob(job *Job) {
	s.Log.Debugf("Handling job: %v", job.JobID)

	ctx, cancel := s.setJobRunning(job)
	defer s.unsetJobRunning(job)

	actor, ok := s.Opts.ActorStorage.Get(job.Node.ActorID)
	if !ok {
		s.Log.Errorf("No such agent: %s", job.Node.ActorID.String())
		return
	}
	if !actor.Online {
		s.Log.Noticef("Agent %s is offline.", actor.Name)
		s.pushback <- job
		return
	}
	nc := tools.NewServerToClientConn(s.Conn, actor.AccountID.String())
	agent := api.NewActorAgentClient(nc, job.Node.ActorID.String())
	req := api.ActorProcessRequest{
		Context: &api.ProcessingContext{
			ProcessID: job.ProcessID,
			NodeID:    job.Node.NodeID,
		},
	}

	if s.processManager.ProcessStorage.GetState(job.ProcessID).Status > api.Process_RUNNING {
		return
	}
	var envelopeIDs []api.UUID
	for _, input := range s.processManager.ProcessStorage.GetInputs(
		job.ProcessID, job.Node,
	) {
		envelope, pos, err := s.Opts.EnvelopeStorage.ReadEnvelope(
			input.EnvelopeID, storage.EnvelopeStartPos(), 100*1024)
		if err != nil {
			s.Log.Errorf("Could not read envelope %s", input.EnvelopeID)
			return
		}
		req.Inputs = append(req.Inputs, &api.ActorProcessRequest_Input{
			Name:     input.Name,
			Envelope: &envelope,
			Position: storage.EncodePosition(envelope.ID, pos),
		},
		)
		envelopeIDs = append(envelopeIDs, envelope.ID)
	}

	if job.attempt == 0 {
		job.firstattempt = time.Now()
	}

	job.Status = api.ActorProcessingState_RUNNING
	job.EnvelopeIDs = envelopeIDs
	s.Opts.JobStorage.SetState(job.JobID, job.JobState)

	// timeoutCount only apply to the current set of attempts
	job.timeoutCount = 0

	watchSub := s.envWatcher.Watch(func(event api.EnvelopeEvent) {
		switch event.Status {
		case api.EnvelopeEvent_STALLED, api.EnvelopeEvent_ERROR:
			// Cancel any current call
			cancel()

			// Cancel the whole process
			process, err := s.Opts.ProcessManager.Get(job.JobID.ProcessID)
			if err != nil {
				s.Log.Errorf("Error getting process %s: %s", job.JobID.ProcessID, err)
			}

			process.Cancel(fmt.Sprintf("Envelope %s status is %s", event.EnvelopeID, event.Status))

		// TODO Mark all the unfinished envelopes as error so they cannot receive
		// more fragments (not here actually, but in the process cancellation handling)
		default:
		}

	}, envelopeIDs...)
	defer watchSub.Unsubscribe()

	for {
		s.Log.Debugf("Calling Process on %s (attempt %d)", actor.Name, job.attempt)
		err := agent.Process(
			ctx, req,
			func(ctx context.Context, state api.ActorProcessingState) {
				job.Status = state.Status
				s.Log.Debugf("Saving job new state: %v", job.JobState)
				if state.ActorLeaving {
					s.actorMap.SetOnline(job.Node.ActorID, false)
				}
				s.Opts.JobStorage.SetState(job.JobID, job.JobState)
				switch state.Status {
				case api.ActorProcessingState_RUNNING:
				default:
					directorServer{s}.ProcessingEnd(ctx, job.Node.ActorID.String(), state)
				}
			},
		)
		s.Log.Noticef("Result of %s.Process: %s", actor.Name, err)

		job.attempt++

		if err == nil {
			if job.Status == api.ActorProcessingState_RUNNING {
				// the job is detached, and will report its status asynchronously
				// with the Director.ProcessingEnd API entry point
				job.Detached = true
				s.Opts.JobStorage.SetState(job.JobID, job.JobState)
			}
			s.Opts.ActorStorage.JustSeen(job.Node.ActorID)
			return
		}

		s.Log.Noticef("Got err: %v", err)
		if err == nats.ErrTimeout {
			job.timeoutCount++
			s.Log.Errorf("Agent %s unresponsive (%d time)", actor.Name, job.timeoutCount)
			if job.timeoutCount == 1 {
				s.Log.Noticef("Marking %s unresponsive", actor.Name)
				s.Opts.ActorStorage.MarkUnresponsive(job.Node.ActorID, job.firstattempt)
			} else if job.timeoutCount >= 5 {
				s.Log.Noticef("Marking %s offline", actor.Name)
				s.Opts.ActorStorage.SetOnline(job.Node.ActorID, false)
				s.pushback <- job
				return
			}
			s.Log.Debugf("Sleeping %d ms",
				time.Duration(job.timeoutCount)*250)
			time.Sleep(time.Duration(job.timeoutCount) * 250 * time.Millisecond)
		} else if nrpcErr, ok := err.(*nrpc.Error); ok && nrpcErr.GetType() == nrpc.Error_SERVERTOOBUSY {
			// Back to pending, it case it is all restarted
			job.Status = api.ActorProcessingState_PENDING
			s.Opts.JobStorage.SetState(job.JobID, job.JobState)
			job.timeoutCount = 0
			delay := (1 << min(job.attempt, 11)) * time.Second
			s.Log.Noticef("Agent %s cannot handle Process() now, wait %s before retrying. Reason given is: %s",
				actor.Name, delay, nrpcErr.Message)
			select {
			case <-ctx.Done():
				return
			case <-time.After(delay):
			}
		} else {
			s.Log.Errorf("Error calling agent %s Process (attempt %d): %s",
				actor.Name, job.attempt, err)
			return
		}

	}
}
