// Copyright 2016 Orus.io. All rights reserved.

package xbus

import (
	"context"
	"crypto"
	"crypto/ecdsa"
	"crypto/tls"
	"crypto/x509"
	"fmt"
	"strings"
	"time"

	"github.com/getsentry/raven-go"
	"github.com/nats-io/go-nats"
	"github.com/nats-io/nuid"
	"github.com/nats-rpc/gogo-nrpc"

	api "bitbucket.org/orus-io/go-xbus/api"
	api_control "bitbucket.org/orus-io/go-xbus/api/control"
	"bitbucket.org/orus-io/go-xbus/tools"
)

var (
	// ClientName can be set by the program to advertise its name to the xbus server
	ClientName string
	// ClientVersion can be set by the program to advertise its version to the xbus server
	ClientVersion string
)

// ServiceOptions holds options for an actor service
type ServiceOptions struct {
	Type     string
	Settings map[string]interface{}
}

// ActorOptions holds options for an actor
type ActorOptions struct {
	api.Actor `yaml:",inline"`
	Service   ServiceOptions
}

// PersistentStore is a key-value store that
type PersistentStore interface {
	Put(key string, value string) error
	Get(key string) (string, error)
	Del(key string) error
}

// TLSOptionsSaver saves tls options
type TLSOptionsSaver interface {
	SetPrivKey(*ecdsa.PrivateKey) error
	SetCertificate([]byte) error
	SetCSR([]byte) error
	SetRootCAs([][]byte) error
}

// TLSOptions regroups the tls configuration entries
type TLSOptions struct {
	PubKey  crypto.PublicKey
	PrivKey *ecdsa.PrivateKey

	Certificate []byte
	CSR         []byte
	RootCAs     [][]byte

	TLSSaver TLSOptionsSaver
}

// SetPrivKey changes the privkey
func (o *TLSOptions) SetPrivKey(key *ecdsa.PrivateKey) error {
	o.PrivKey = key
	o.PubKey = o.PrivKey.Public()
	if o.TLSSaver != nil {
		return o.TLSSaver.SetPrivKey(key)
	}
	return nil
}

// SetCertificate changes the certificate
func (o *TLSOptions) SetCertificate(cert []byte) error {
	o.Certificate = cert
	if o.TLSSaver != nil {
		return o.TLSSaver.SetCertificate(cert)
	}
	return nil
}

// SetCSR changes the csrificate
func (o *TLSOptions) SetCSR(csr []byte) error {
	o.CSR = csr
	if o.TLSSaver != nil {
		return o.TLSSaver.SetCSR(csr)
	}
	return nil
}

// SetRootCAs changes the root cas
func (o *TLSOptions) SetRootCAs(cas [][]byte) error {
	o.RootCAs = cas
	if o.TLSSaver != nil {
		return o.TLSSaver.SetRootCAs(cas)
	}
	return nil
}

// COptions configures a Client
type COptions struct {
	ID   api.UUID
	Name string
	Type api.Account_Type

	NatsTimeout time.Duration

	ForceStart bool

	ActorList []ActorOptions

	ConnOptions nats.Options

	PersistentStore PersistentStore

	SentryDSN string

	TLS TLSOptions
}

// NewClient creates a new client
func NewClient(options COptions, logger Logger) *Client {
	if options.ID.IsZero() && options.Name == "" {
		// TODO NewClient should return an error
		panic("Account ID and name both missing")
	}
	if options.PersistentStore == nil {
		var err error
		options.PersistentStore, err = NewFilePersistentStore("pstore.db")
		if err != nil {
			panic(err)
		}
	}
	var ravenClient *raven.Client
	if options.SentryDSN != "" {
		var err error
		ravenClient, err = raven.New(options.SentryDSN)
		if err != nil {
			panic(err)
		}
		logger = NewRavenLogger(ravenClient, logger)
		logger.Noticef("Errors will be sent to sentry")
	}
	return &Client{
		options: options,
		Log:     logger,
		actors:  make(map[api.UUID]*Actor),
		stopped: make(chan bool),

		processingEndQueue:      make(chan processingEndEntry, 1),
		processingEndForceRetry: make(chan bool),

		ravenClient: ravenClient,
	}
}

// processingEndEntry
type processingEndEntry struct {
	actorID  api.UUID
	context  api.ProcessingContext
	status   api.ActorProcessingState_Status
	messages []api.LogMessage
}

// Client is a xbus client api
type Client struct {
	options       COptions
	conn          *nats.Conn
	authenticated bool
	Log           Logger
	actors        map[api.UUID]*Actor
	running       []*RunningService
	quit          chan bool
	stopped       chan bool

	processingEndQueue       chan processingEndEntry
	processingEndForceRetry  chan bool
	processingEndWaitingList []processingEndEntry

	ravenClient *raven.Client

	API     *api.Client
	Control *api_control.Client
}

// GetOptions returns the current client options
func (c *Client) GetOptions() COptions {
	return c.options
}

// GetConn returns the underlying nats connection
func (c *Client) GetConn() *nats.Conn {
	return c.conn
}

// GetPersistentStore returns the persistentstore
func (c *Client) GetPersistentStore() PersistentStore {
	return c.options.PersistentStore
}

// ID returns the account ID (may be zero)
func (c *Client) ID() api.UUID {
	return c.options.ID
}

// Name returns the account Name
func (c *Client) Name() string {
	return c.options.Name
}

// ActorOptions returns the list of all the actors
func (c *Client) ActorOptions() []ActorOptions {
	return c.options.ActorList
}

// Connect establises a connection to the bus
func (c *Client) Connect() error {
	return c.connect(nil)
}

func (c *Client) connect(disconnectedCB nats.ConnHandler) error {
	var (
		anon = c.options.TLS.Certificate == nil
		ns   string
	)
	opts := c.options.ConnOptions
	opts.Secure = true
	opts.DisconnectedCB = disconnectedCB
	opts.Name = fmt.Sprintf("%s@%s", c.options.Name, api.ProtocolVersion)

	if ClientName != "" || ClientVersion != "" {
		opts.Name = fmt.Sprintf("%s(%s@%s)", opts.Name, ClientName, ClientVersion)
	}

	if anon {
		c.Log.Noticef("Anonymous connection")
		opts.User = nuid.Next()
		ns = "anon-" + opts.User

	} else {
		c.Log.Debugf("Using TLS cert to connect")

		// Fill TLSConfig elements based on upper settings when necessary.
		if opts.TLSConfig == nil {
			opts.TLSConfig = &tls.Config{}
		}
		if opts.TLSConfig.Certificates == nil {
			opts.TLSConfig.Certificates = []tls.Certificate{tls.Certificate{
				Certificate: [][]byte{c.options.TLS.Certificate},
				PrivateKey:  c.options.TLS.PrivKey,
			}}
		}
		if opts.TLSConfig.RootCAs == nil {
			var rootCAs *x509.CertPool
			if c.options.TLS.RootCAs != nil {
				rootCAs = x509.NewCertPool()
				for _, ca := range c.options.TLS.RootCAs {
					cert, err := x509.ParseCertificate(ca)
					if err != nil {
						return fmt.Errorf("Cannot parse root ca: %s", err)
					}
					rootCAs.AddCert(cert)
				}
			}
			opts.TLSConfig.RootCAs = rootCAs
		}
		ns = c.options.ID.String()
	}

	conn, err := opts.Connect()
	if err != nil {
		return err
	}

	c.conn = conn

	c.authenticated = c.options.TLS.Certificate != nil

	if !anon {
		// We need to get more info about the client account
		id := c.options.Name
		if c.options.Type == api.Account_GATEWAY {
			switch {
			case c.options.ConnOptions.Token == "gateway":
			case strings.HasPrefix(c.options.ConnOptions.Token, "webhook:"):
				id = c.options.ConnOptions.Token[8:]
			default:
				id = c.options.ConnOptions.Token
			}
		}
		_, err := c.WhoAmI(id)
		if err != nil {
			c.conn = nil
			conn.Close()
			return err
		}

	} else {
		c.API = api.NewClient(tools.NewClientConn(conn, ns))
		c.Control = api_control.NewClient(tools.NewClientConn(conn, ns))

		if c.options.NatsTimeout != 0 {
			c.API.SetTimeout(c.options.NatsTimeout)
			c.Control.SetTimeout(c.options.NatsTimeout)
		}
	}

	if !c.options.ID.IsZero() && c.options.Type == api.Account_ACTOR {
		// update the actor definitions
		c.UpdateActors()
	}

	return nil
}

// Close closes the connection to the bus
func (c *Client) Close() {
	c.conn.Close()
	c.conn = nil
}

// AddActor attach an actor
func (c *Client) AddActor(actor ActorOptions) {
	c.options.ActorList = append(c.options.ActorList, actor)
}

// GetActorOptions returns an actor options from its id
func (c *Client) GetActorOptions(id api.UUID) (ActorOptions, bool) {
	for _, actor := range c.options.ActorList {
		if actor.ID == id {
			return actor, true
		}
	}
	return ActorOptions{}, false
}

// GetActorOptionsByName returns an actor options from its name
func (c *Client) GetActorOptionsByName(name string) (ActorOptions, bool) {
	for _, actor := range c.options.ActorList {
		if actor.Name == name {
			return actor, true
		}
	}
	return ActorOptions{}, false
}

// getActorFromOptions returns a *Actor from the options
func (c *Client) getActorFromOptions(options ActorOptions) *Actor {
	actor, ok := c.actors[options.ID]
	if !ok {
		actor = newActor(c, options)
		c.actors[options.ID] = actor
	}
	return actor
}

// GetActor returns a *Actor from its ID, or nil if it undefined
func (c *Client) GetActor(id api.UUID) *Actor {
	actor, ok := c.actors[id]
	if !ok {
		options, ok := c.GetActorOptions(id)
		if !ok {
			return nil
		}
		actor = c.getActorFromOptions(options)
	}
	return actor
}

// GetActorByName returns an actor from its name
func (c *Client) GetActorByName(name string) *Actor {
	actorOpts, ok := c.GetActorOptionsByName(name)
	if ok {
		actor, ok := c.actors[actorOpts.ID]
		if !ok {
			actor = c.getActorFromOptions(actorOpts)
		}
		return actor
	}
	return nil
}

type replyType interface {
	GetError() *nrpc.Error
}

// RawProcessingEnd calls ProcessingEnd server API directly
func (c *Client) RawProcessingEnd(
	actorID api.UUID,
	context api.ProcessingContext,
	status api.ActorProcessingState_Status,
	messages []api.LogMessage,
) error {
	return c.API.Director.ProcessingEnd(
		actorID.String(),
		api.ActorProcessingState{
			Context:  context,
			Status:   status,
			Messages: messages,
		},
	)
}

func (c *Client) tryProcessingEnd(
	actorID api.UUID,
	context api.ProcessingContext,
	status api.ActorProcessingState_Status,
	messages []api.LogMessage,
	firstAttempt bool,
) (tryAgain bool, err error) {
	key := actorID.String() +
		"/" + context.ProcessID.String() +
		"/" + context.NodeID +
		"/" + "processingend"

	if firstAttempt {
		sStatus, err := status.MarshalText()
		if err != nil {
			panic(err)
		}
		if err := c.GetPersistentStore().Put(key, string(sStatus)); err != nil {
			c.Log.Errorf("Error using persistent store: %s.", err)
		}
	}

	if err := c.RawProcessingEnd(actorID, context, status, messages); err != nil {
		if err == nats.ErrTimeout ||
			err == nats.ErrConnectionClosed ||
			err == nats.ErrNoServers ||
			err == nats.ErrStaleConnection {
			// Most probably a temporary error (broker is down)
			// We will try again later in the background
			if firstAttempt {
				c.processingEndQueue <- processingEndEntry{
					actorID, context, status, messages,
				}
			}
			return true, err
		}
		return false, err
	}
	c.GetPersistentStore().Del(key)
	return false, nil
}

func (c *Client) processingEndWorker() {
	retry := make(chan bool)
	delay := time.Second

	doRetry := func() {
		if len(c.processingEndWaitingList) == 0 {
			return
		}
		var stillWaiting []processingEndEntry
		c.Log.Noticef("Attempt to send waiting ProcessingEnd(s)")
		for i, entry := range c.processingEndWaitingList {
			tryAgain, err := c.tryProcessingEnd(
				entry.actorID, entry.context, entry.status,
				entry.messages, false)
			if err != nil {
				c.Log.Errorf("Error calling ProcessingEnd: %s", err)
			} else {
				delay = time.Second
			}
			if tryAgain {
				stillWaiting = c.processingEndWaitingList[i:]
				if delay < 5*time.Minute {
					delay = delay * 16 / 10
				}
				break
			}
		}
		if len(stillWaiting) != 0 {
			c.Log.Errorf("Will try again in '%f' seconds", delay.Seconds())
			time.AfterFunc(delay, func() { retry <- true })
		}
	}

	for {
		select {
		case newEntry := <-c.processingEndQueue:
			if len(c.processingEndWaitingList) == 0 {
				time.AfterFunc(delay, func() { retry <- true })
			}
			c.processingEndWaitingList = append(c.processingEndWaitingList, newEntry)
		case <-retry:
			doRetry()
		case <-c.processingEndForceRetry:
			delay = time.Second
			doRetry()
		case <-c.quit:
			return
		}
	}
}

type actorServiceTuple struct {
	actor *Actor
	svc   ActorService
}

// StartServices initialize and starts all the actors having a ActorService
func (c *Client) StartServices() error {
	var actorServices []actorServiceTuple
	for _, actorOpts := range c.options.ActorList {
		if actorOpts.Service.Type != "" {
			actor := c.getActorFromOptions(actorOpts)
			actorService := MakeActorService(actorOpts.Service.Type, actor)
			if actorService == nil {
				return fmt.Errorf(
					"No such actor service: %s", actorOpts.Service.Type)
			}
			actorServices = append(actorServices, actorServiceTuple{actor, actorService})
		}
	}
	for _, as := range actorServices {
		runningSvc, err := StartupService(as.actor, as.svc)
		if err != nil {
			c.Log.Errorf("Error starting actor %s: %s", as.actor.Name, err)
			if err := c.StopServices(); err != nil {
				c.Log.Errorf("Error shutting down actors: %s", err)
			}
			return err
		}
		c.running = append(c.running, runningSvc)
	}
	return nil
}

// StopServices stops all the running actor
func (c *Client) StopServices() error {
	var lastError error
	actors := c.running[:]
	for _, actor := range actors {
		c.running = c.running[1:]
		if lastError = actor.Shutdown(); lastError != nil {
			c.Log.Errorf("Error shutting down actor: %s", lastError)
		}
	}
	return lastError
}

// Startup attempts to connect and start the services.
// If the connection attempt fail, it waits a little and
// retries forever.
// It blocks until a first successful connect + services start is performed*
func (c *Client) Startup(startServices bool) error {
	var (
		disconnected = make(chan bool)
		retry        = make(chan bool)
	)
	if c.ravenClient != nil {
		defer c.ravenClient.Wait()
	}
	err := c.connect(func(*nats.Conn) {
		disconnected <- true
	})
	if err != nil {
		if !c.options.ForceStart {
			return err
		}
		time.AfterFunc(time.Second, func() { retry <- true })
	} else {
		if err := c.StartServices(); err != nil {
			return err
		}
	}

	c.quit = make(chan bool)

	go c.processingEndWorker()

	go func() {
		defer func() {
			c.stopped <- true
		}()
		for {
			select {
			case <-disconnected:
				c.Log.Errorf("Got disconnected. Stopping services then reconnect")
				if err := c.StopServices(); err != nil {
					c.Log.Errorf("Error stopping services: %s", err)
				}
				c.Close()
			case <-c.quit:
				if err := c.StopServices(); err != nil {
					c.Log.Errorf("Error stopping services: %s", err)
				}
				if c.conn != nil {
					c.Close()
				}
				return
			case <-retry:
			}
			c.Log.Debugf("Reconnecting")
			if err := c.connect(func(*nats.Conn) {
				disconnected <- true
			}); err != nil {
				c.Log.Errorf("Error reconnecting: %s", err)
				time.AfterFunc(time.Second, func() { retry <- true })
				continue
			}
			c.Log.Debugf("Start services")
			if startServices {
				err := c.StartServices()
				if err != nil {
					c.Log.Errorf("Error starting services: %s", err)
					c.Close()
					time.AfterFunc(time.Second, func() { retry <- true })
				}
			}
			c.processingEndForceRetry <- true
		}
	}()
	return nil
}

// Shutdown stops the services if running and disconnect from
// the bus
func (c *Client) Shutdown() {
	close(c.quit)
	<-c.stopped
}

// ClientGetActors lists actors of the currently connected account
func (c *Client) ClientGetActors() ([]api.Actor, error) {
	rep, err := c.API.NewClientAPI(c.ID().String()).GetActors()
	return rep.GetActors(), err
}

// UpdateActors gets the actors definitions from the server and merge then with
// current actors
func (c *Client) UpdateActors() error {
	actors, err := c.ClientGetActors()
	if err != nil {
		return err
	}

	index := make(map[string]api.Actor)
	for _, actor := range actors {
		index[actor.Name] = actor
	}

	for i, actor := range c.options.ActorList {
		if actorDef, ok := index[actor.Name]; ok {
			c.options.ActorList[i].Actor = actorDef
			delete(index, actor.Name)
		}
	}

	for _, actor := range index {
		c.AddActor(ActorOptions{Actor: actor})
	}

	return nil
}

// WhoAmI returns informations on the currently connect account
// Works only when connected with a API Key
func (c *Client) WhoAmI(apiKey string) (api.Account, error) {
	whoami := api.NewWhoAmIClient(tools.NewWhoAmIConn(c.conn))
	account, err := whoami.WhoAmI(apiKey)
	if err != nil {
		return account, err
	}
	c.options.ID = account.ID
	c.options.Name = account.Name
	c.options.Type = account.Type

	ns := c.options.ID.String()
	c.API = api.NewClient(tools.NewClientConn(c.conn, ns))
	c.Control = api_control.NewClient(tools.NewClientConn(c.conn, ns))

	if c.options.NatsTimeout != 0 {
		c.API.SetTimeout(c.options.NatsTimeout * time.Second)
		c.Control.SetTimeout(c.options.NatsTimeout * time.Second)
	}

	return account, nil
}

// RetrieveEnvelope fetch an envelope from the EnvelopeStorage service and
// returns EnvelopeReceiver for reading it on the fly
// The fetching can be stopped by canceling the context
func (c *Client) RetrieveEnvelope(
	ctx context.Context, clientKind, clientID string, envelopeID api.UUID,
) *EnvelopeReceiver {
	receiver := NewEnvelopeReceiverSimple(envelopeID)
	go func() {
		myCtx, cancel := context.WithCancel(ctx)
		if err := c.API.EnvelopeStorage.Retrieve(
			myCtx, clientKind, clientID, api.EnvelopePosition{EnvelopeID: envelopeID, Start: true},
			func(ctx context.Context, fragmentAndPosition api.EnvelopeFragmentAndPosition) {
				if err := receiver.AddFragment(*fragmentAndPosition.Fragment); err != nil {
					cancel()
					receiver.SetError(err)
				}
			},
		); err != nil {
			receiver.SetError(err)
		} else {
			if err := receiver.Finalize(); err != nil {
				c.Log.Errorf("Error finalizing the envelope receiver: %s", err)
			}
		}
	}()
	return receiver
}

// RetrieveEnvelopeFrom retrieves an envelope starting at a given position
func (c *Client) RetrieveEnvelopeFrom(
	ctx context.Context, clientKind, clientID string,
	envelope api.Envelope, position api.EnvelopePosition,
) *EnvelopeReceiver {
	receiver, _ := NewEnvelopeReceiver("", envelope)
	if receiver.Status() == api.EnvelopeAck_RECEIVING {
		go func() {
			myCtx, cancel := context.WithCancel(ctx)
			receiver.OnCancel = cancel
			if err := c.API.EnvelopeStorage.Retrieve(
				myCtx, clientKind, clientID, position,
				func(ctx context.Context, fragmentAndPosition api.EnvelopeFragmentAndPosition) {
					if err := receiver.AddFragment(*fragmentAndPosition.Fragment); err != nil {
						cancel()
						receiver.SetError(err)
					}
				},
			); err != nil {
				receiver.SetError(err)
			} else {
				if err := receiver.Finalize(); err != nil {
					c.Log.Errorf("Error finalizing the envelope receiver: %s", err)
				}
			}
		}()
	}
	return receiver
}
